/*
 * $Id: s.cfggsc.c 1.1 96/12/29 19:49:33-00:00 mjacob@bird.feral.com $
 * Copyright (c) 1997 by Matthew Jacob
 * Copyright (c) 2011 Gresham Enterprise Storage
 *
 *	This software is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Library General Public
 *	License as published by the Free Software Foundation; version 2.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public
 *	License along with this software; if not, write to the Free
 *	Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *	The author may be reached via electronic communications at
 *
 *		mjacob@feral.com
 *
 *	or, via United States Postal Address
 *
 *		Matthew Jacob
 *		1831 Castro Street
 *		San Francisco, CA, 94131
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/scsi.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/devinfo.h>
#include <sys/device.h>
#include <sys/cfgodm.h>
#include <cf.h>
#include <sys/cfgdb.h>
#include <fcntl.h>
#include <sys/sysconfig.h>
#include <sys/errno.h>
#include <sys/device.h>
#include <sys/mode.h>

#include "gscdds.h"
#include "gscdds_internal.h"

static int verbose;
static struct gsc_ddsinfo ddsinfo;
#ifdef FCP
static struct gsc_qvpd qvpd;
#endif

int main(int a, char **v);

#define MKNOD_MODE S_IFCHR|S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH

#define	vprintf	if (verbose) printf

#ifndef FCP
static int get_scsi_lun(const char *str)
{
	char conns[256];
	char *ptr;
	int lun;

	/* Format is like "10-60-00-4,0", where scsi id is 4 and lun is 0  */
	strncpy(conns, str, sizeof(conns));
	conns[sizeof(conns)-1] = 0;
    vprintf("Extract scsi_id/lun from location %s\n", conns);

	/* Get lun */
    ptr = conns;
    while (*ptr && ptr < &conns[NAMESIZE])
		ptr++;
    ptr--;
    if (ptr < &conns[1]) {
		return E_BADATTR;
    }
    lun = *ptr - '0';
    vprintf("I see lun %d\n", lun);
    ddsinfo.lun_id = lun;

    /* Get scsi id */
    if (ptr[-1] == ',') {
		*(--ptr) = 0;
    } else {
		*ptr = 0;
    }
    while (ptr > conns && *ptr != '-')
		ptr--;
    if (*ptr == '-')
		ptr++;
    
    ddsinfo.scsi_id = strtol(ptr, (char **) NULL, 0);

    if (lun < 0 || lun >= 8)
		return E_INVCONNECT;

	return 0;
}
#endif

static void
err_exit(char exitcode)
{
    odm_close_class(CuDv_CLASS);
    odm_close_class(PdDv_CLASS);
    odm_terminate();
    exit(exitcode);
}

int
main(int argc, char *argv[])
{
    char sstring[256], bufname[256], *ptr, *lname;
    struct Class *cprp, *cusdev, *predev;
    struct CuDvDr dpr;
    struct CuDv parobj, cusobj;
    struct PdDv preobj;
    int rc, c, errflg, ipl_phase, unit, howmany, domake;
    long major;
    int *mlist;
    mid_t kmid;
    struct cfg_dd dd;
#ifdef FCP
	struct CuAt *CuAt;
	int how_many;
#endif

    lname = NULL;
    errflg = 0;
    ipl_phase = RUNTIME_CFG;
    (void) memset((void *) &ddsinfo, 0, sizeof ddsinfo);
	
    while ((c = getopt(argc, argv, "l:12v")) != EOF) {
		switch (c) {
		case 'l':
			if (lname != NULL)
				errflg++;
			lname = optarg;
			break;
		case 'v':
			verbose++;
			break;
		case '1':
			if (ipl_phase != RUNTIME_CFG)
				errflg++;
			ipl_phase = PHASE1;
			break;
		case '2':
			if (ipl_phase != RUNTIME_CFG)
				errflg++;
			ipl_phase = PHASE2;
			break;
		default:
			errflg++;
		}
    }
    if (errflg)
		exit (E_ARGS);
    if (lname == NULL)
		exit(E_LNAME);

    if (odm_initialize() == -1) {
		return (E_ODMINIT);
    }

    /* lock the database */
    if (odm_lock("/etc/objrepos/config_lock",0) == -1)
		err_exit(E_ODMLOCK);
	
    /* open customized devices object class */
    if ((int)(cusdev = odm_open_class(CuDv_CLASS)) == -1)
		err_exit(E_ODMOPEN);

    /* search for customized object with this logical name */
    sprintf(sstring, "name = '%s'", lname);
    rc = (int) odm_get_first(cusdev, sstring, &cusobj);
    if (rc == 0) {
		/* No CuDv object with this name */
		err_exit(E_NOCuDv);
    } else if (rc == -1) {
		/* ODM failure */
		err_exit(E_ODMGET);
    }

    /* open predefined devices object class */
    if ((int)(predev = odm_open_class(PdDv_CLASS)) == -1)
		err_exit(E_ODMOPEN);
	
    /* get predefined device object for this logical name */
    sprintf(sstring, "uniquetype = '%s'", cusobj.PdDvLn_Lvalue);
    rc = (int)odm_get_first(predev, sstring, &preobj);
    if (rc == 0) {
		/* No PdDv object for this device */
		err_exit(E_NOPdDv);
    } else if (rc == -1) {
		/* ODM failure */
		err_exit(E_ODMGET);
    }
    /* close predefined device object class */
    if (odm_close_class(predev) == -1)
		err_exit(E_ODMCLOSE);

    if (ipl_phase != RUNTIME_CFG)
		setleds(preobj.led);

    /*
     * Now, if the device is already configured, we're
     * pretty much done.
     */
    if (cusobj.status == AVAILABLE) {
		/* close customized device object class */
		if (odm_close_class(cusdev) == -1)
			err_exit(E_ODMCLOSE);
		odm_terminate();
		return(E_OK);
    }
    if (cusobj.status != DEFINED) {
		vprintf("bad state: %d\n", cusobj.status);
		err_exit(E_DEVSTATE);
    }

    /* get the device's parent object */
    sprintf(sstring, "name = '%s'", cusobj.parent);
    rc = (int) odm_get_first(cusdev, sstring, &parobj);
    if (rc == 0) {
		/* Parent device not in CuDv */
		err_exit(E_NOCuDvPARENT);
    } else if (rc == -1) {
		/* ODM failure */
		err_exit(E_ODMGET);
    }
	
    /* Parent MUST be available to continue */
    if (parobj.status != AVAILABLE)
		err_exit(E_PARENTSTATE);

    /* make sure that no other devices are configured     */
    /* at this location                                   */
#ifdef FCP
    sprintf(sstring, "parent = '%s' AND location='%s' AND connwhere='%s' AND status=%d",
			cusobj.parent, cusobj.location, cusobj.connwhere, AVAILABLE);
#else
	sprintf(sstring, "parent = '%s' AND location='%s' AND status=%d",
			cusobj.parent, cusobj.location, AVAILABLE);
#endif
    rc = (int) odm_get_first(cusdev, sstring, &cusobj);
    if (rc == -1) {
		/* odm failure */
		err_exit(E_ODMGET);
    } else if (rc) {
		/* Error: device config'd at this location */
		err_exit(E_AVAILCONNECT);
    }

#ifdef FCP
	CuAt = getattr(lname, "scsi_id", 0, &how_many);
	if (!CuAt) {
		fprintf(stderr, "scsi_id");
		err_exit(E_NOATTR);
	}

	rc = sscanf(CuAt->value, "%llx", &ddsinfo.scsi_id);
	if (rc != 1) {
		fprintf(stderr, "scsi_id");
		err_exit(E_BADATTR);
	}

	CuAt = getattr(lname, "lun_id", 0, &how_many);
	if (!CuAt) {
		fprintf(stderr, "lun_id");
		err_exit(E_NOATTR);
	}

	rc = sscanf(CuAt->value, "%llx", &ddsinfo.lun_id);
	if (rc != 1) {
		fprintf(stderr, "lun_id");
		err_exit(E_BADATTR);
	}
#else
	rc = get_scsi_lun(cusobj.location);

	if (rc) {
		err_exit(rc);
	}
#endif

    vprintf("I see tgt 0x%llx, lun 0x%llx\n", ddsinfo.scsi_id, ddsinfo.lun_id);

    /*
     * Generate dev_t for adapter
     */
    cprp = odm_open_class(CuDvDr_CLASS) ;
    sprintf(sstring, "value3 = %s", cusobj.parent);
    rc = (int) odm_get_obj(cprp, sstring, &dpr, TRUE);
    if (rc == 0) {
		err_exit(E_NOCuDvPARENT);
    } else if (rc == -1) {
		err_exit(E_ODMGET);
    }
    ddsinfo.busid = (dev_t) makedev(atoi(dpr.value1), atoi(dpr.value2));
    vprintf("I see %d.%d for connecting adapter\n",
			major(ddsinfo.busid), minor(ddsinfo.busid));

    /*
     * Get unit number out of logical name
     */
	
    ptr = lname;
    ptr += strlen(preobj.prefix);
    unit = atoi(ptr);
    vprintf("I see %d as unit\n", unit);
	
    /*
     * Okay, now that we have the pertinent information that we'll
     * need (adapter dev_t, device type, target, lbits, shareable,
     * unit number), we can look into actually loading/configuring the
     * current driver.
     */
    (void) sprintf(bufname, "/dev/%s", lname);
	
    /*
     * Get or generate major number..
     */
    if ((major = (long) genmajor(preobj.DvDr)) == -1) {
		odm_terminate();
		return (E_MAJORNO);
    }
    vprintf("major is %ld\n", major);

    /*
     * Let's see if this is the first time through. If it's
     * the first time through, getminor will return NULL
     * or won't have any minors in the list.
     */
    mlist = getminor(major, &howmany, preobj.DvDr);
    vprintf("getminor: %x and howmany %d for %s\n", *mlist, howmany,
			preobj.DvDr);
	
    domake = 1;
    if (mlist != NULL && howmany != 0) {
		/*
		 * We have a list of minors already.
		 * See if we already have the minor
		 * we want defined.
		 */
		for (c = 0; c < howmany; c++) {
			if (mlist[c] == unit) {
				vprintf("unit %d already has minor\n", unit);
				domake = 0;
				break;
			}
		}
    }

    if (domake) {
		(void) unlink(bufname);
		/*
		 * Now create the minor number that will match the unit number.
		 * We really don't care whether genminor succeeds, since
		 * we've alreay unlinked the device node.
		 */
		mlist = genminor(preobj.DvDr, major, unit, 1, 1, 1);
		if (mlist == NULL) {
			vprintf("genminor failed\n");
			err_exit(E_MINORNO);
		}
		vprintf("making %s as %ld.%d with minor returned as %d\n",
				bufname, major, unit, *mlist);
		if (mknod(bufname, MKNOD_MODE, makedev(major, unit))) {
			vprintf("Cannot make special device\n");
			err_exit(E_MKSPECIAL);
		}
    } else {
		mknod(bufname, MKNOD_MODE, makedev(major, unit));
    }
	
    /*
     * Load the driver....
     */
    kmid = loadext(preobj.DvDr, TRUE, FALSE);
    if (!kmid) {
		vprintf("Cannot load the driver [%s]\n", preobj.DvDr);
		err_exit(E_LOADEXT);
    }
	
    /*
     * And configure the driver...
     */
	strncpy(ddsinfo.devname, lname, sizeof(ddsinfo.devname));
    dd.kmid = kmid;
    dd.devno = makedev(major, unit);
    dd.cmd = CFG_INIT;
    dd.ddsptr = (caddr_t) &ddsinfo;
    dd.ddslen = sizeof (ddsinfo);

    if (sysconfig(SYS_CFGDD, &dd, sizeof (dd)) == CONF_FAIL) {
		int saverr = errno;
		vprintf("sysconfig failed, error is %d\n", errno);

		/*
		 * Unload driver...
		 */
		(void) loadext(preobj.DvDr, FALSE, FALSE);
		switch(saverr) {
		case ENODEV:
			err_exit(E_WRONGDEVICE);
			/* NOTREACHED */
			break;
		case EBUSY:
			err_exit(E_AVAILCONNECT);
			/* NOTREACHED */
			break;
		case EINVAL:
		default:
			err_exit(E_CFGINIT);
			/* NOTREACHED */
			break;
		}
    }

#ifdef FCP
	/* Retrieve the World Wide Name and Node Name of that device. */
    dd.cmd = CFG_QVPD;
    dd.ddsptr = (caddr_t) &qvpd;
    dd.ddslen = sizeof (qvpd);

    if (sysconfig(SYS_CFGDD, &dd, sizeof (dd)) == 0) {

		/* Update node name and WW name. */
		CuAt = getattr(lname, "node_name", 0, &how_many);
		if (CuAt) {
			strcpy(CuAt->value, qvpd.node_name);
			putattr(CuAt);
		}
		CuAt = getattr(lname, "ww_name", 0, &how_many);
		if (CuAt) {
			strcpy(CuAt->value, qvpd.ww_name);
			putattr(CuAt);
		}
	}
#endif
	
    /* now mark the device as available */
    cusobj.status = AVAILABLE;
    if (odm_change_obj(CuDv_CLASS, &cusobj) == -1) {
		/*
		 * Unconfigure driver (for this instance)...
		 */
		dd.kmid = 0;
		dd.ddsptr = (caddr_t) NULL;
		dd.ddslen = (int ) 0;
		dd.cmd = CFG_TERM;
		(void) sysconfig(SYS_CFGDD, &dd, sizeof (dd));
		/*
		 * Unload driver...
		 */
		(void) loadext(preobj.DvDr, FALSE, FALSE);
		err_exit (E_ODMUPDATE);
    }
	
    (void) odm_terminate();
    return (E_OK);
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * End:
 */
